import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:rxdart/rxdart.dart';

import 'property_state.dart';

/// Base class for business logic components.
///
/// This is inspired by the BLoC pattern but not quite the same.
/// A bloc provides methods that trigger asynchronous procedures
/// which in turn publish state changes through stream controllers.
///
/// A bloc can provide multiple streams to the UI layer.
abstract class BaseBloc {
  final _streamControllers = <StreamController>[];
  final _streamSubscriptions = <StreamSubscription>[];

  BaseBloc();

  void initialize();

  /// Instantiates an auto-disposing [BehaviourSubject]
  ///
  /// Disposal is taken care of in the [dispose] method.
  BehaviorSubject<TState> subject<TState>() {
    final subject = BehaviorSubject<TState>();
    _streamControllers.add(subject);
    return subject;
  }

  /// Instantiates an auto-disposing [BehaviourSubject] with
  /// a [PropertyState].
  ///
  /// Disposal is taken care of in the [dispose] method.
  BehaviorSubject<PropertyState<TState>> property<TState>(
      [PropertyState<TState>? initial]) {
    final controller = subject<PropertyState<TState>>();
    if (initial != null) {
      controller.sink.add(initial);
    }
    return controller;
  }

  /// Instantiates an auto-disposing [StreamController].
  ///
  /// Disposal is taken care of in the [dispose] method.
  StreamController<TState> stream<TState>() {
    final subject = StreamController<TState>();
    _streamControllers.add(subject);
    return subject;
  }

  /// Register a [StreamSubscription] for auto-disposal.
  ///
  /// Disposal is taken care of in the [dispose] method.
  //TODO better name for this method
  void sub(StreamSubscription subscription) {
    _streamSubscriptions.add(subscription);
  }

  @mustCallSuper
  void dispose() {
    // close all state stream controllers
    _streamControllers.forEach((s) {
      s.close();
    });
    // cancel all prerecorded stream subscriptions
    _streamSubscriptions.forEach((s) {
      s.cancel();
    });
  }
}
