import 'package:boost/boost.dart';
import 'package:cl_appbase/src/bloc/reactive_collection.dart';
import 'package:flutter/material.dart';

import 'collection_builder.dart';

class SliverCollectionBuilder<TValue> extends StatefulWidget {
  final ScrollPhysics? physics;
  final Axis scrollDirection;
  final ScrollBehavior? scrollBehavior;

  final ReactiveCollection<TValue> collection;
  final List<Widget>? leading;
  final Widget Function(BuildContext context, TValue) itemBuilder;
  final Widget Function(BuildContext context)? loadingBuilder;
  final Widget Function(BuildContext context)? emptyBuilder;
  final Widget Function(BuildContext context, dynamic e)? errorBuilder;

  const SliverCollectionBuilder({
    Key? key,
    required this.collection,
    required this.itemBuilder,
    this.leading,
    this.errorBuilder,
    this.emptyBuilder,
    this.loadingBuilder,
    this.physics,
    this.scrollDirection = Axis.vertical,
    this.scrollBehavior,
  }) : super(key: key);

  @override
  _SliverCollectionBuilderState<TValue> createState() =>
      _SliverCollectionBuilderState<TValue>();
}

class _SliverCollectionBuilderState<TValue>
    extends State<SliverCollectionBuilder<TValue>> {
  final _continueSemaphore = Semaphore();
  late ScrollController _controller;
  bool _done = false;
  final key = GlobalKey();

  @override
  void initState() {
    super.initState();
    _controller = ScrollController()..addListener(_onScroll);
    _continueLoading();
  }

  @override
  Widget build(BuildContext context) {
    return CollectionBuilder<TValue>(
      key: key,
      stream: widget.collection.stream,
      builder: _build,
      error: widget.errorBuilder ?? _fallbackError,
    );
  }

  Widget _build(BuildContext context, List<TValue> items, bool loading) {
    return CustomScrollView(
      physics: widget.physics,
      controller: _controller,
      scrollDirection: widget.scrollDirection,
      scrollBehavior: widget.scrollBehavior,
      slivers: [
        if (widget.leading != null) ...widget.leading!,
        ...items.map((e) => widget.itemBuilder(context, e)),
        if (loading)
          widget.loadingBuilder?.call(context) ??
              SliverToBoxAdapter(
                child: SizedBox(
                  height: 64.0,
                  child: Center(
                    child: SizedBox(
                      child: SizedBox(
                        height: 32.0,
                        width: 32.0,
                        child: CircularProgressIndicator.adaptive(),
                      ),
                    ),
                  ),
                ),
              )
      ],
    );
  }

  void _onScroll() {
    if (!_done &&
        _controller.offset > _controller.position.maxScrollExtent - 300) {
      _continueLoading();
    }
  }

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }

  void _continueLoading() {
    _continueSemaphore.throttle(() async {
      _done = await widget.collection.load();
    });
  }

  Widget _fallbackError(BuildContext context, dynamic e) {
    return Text(e.toString());
  }
}
